#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
import logging
import tkinter as tk
from tkinter import *
from tkinter import ttk
from tkinter.ttk import *

from geofinder import CachedDictionary, AppStyle
from geofinder import TKHelper as Widge

BUTTON_WIDTH = 6
# tags to aletrnate colors in list box
odd_tag = ('odd',)
even_tag = ('even',)

default = []

class ListboxFrame:
    """
     ListboxFrame -  GUI to display and remove lines from a Cached Dictionary
                     Displays a scrolling list box of data with read/write from a CachedDictionary
                     Allows user to delete items from list
                     Defines overall grid layout for derived classes
    """

    def __init__(self, frame, title, dir_name, cache_filename):
        self.logger = logging.getLogger(__name__)
        self.grd = {"title_label": [0, 0, 5, 5, "EW"],
                    "listbox": [0, 1, 5, 5, "EW"], "scrollbar": [1, 1, 0, 5, "WNS"],
                    "status": [0, 2, 5, 5, "EW"], "load_button": [2, 2, 5, 5, "W"], "remove_button": [2, 2, 5, 5, "W"],
                    "pad": [0, 3, 5, 5, "EW"],
                    "add_label": [0, 4, 5, 5, "EW"], "add_button": [2, 4, 5, 5, "W"],
                    "add_entry": [0, 5, 5, 5, "EW"], "listbox_all_countries": [0, 5, 5, 5, "EW"], "scrollbar2": [1, 5, 0, 5, "WNS"],
                    "country_label": [0, 5, 5, 5, "EW"],
                    "country_entry": [0, 6, 5, 5, "W"], "country_button": [2, 6, 5, 5, "W"],
                    }

        self.title = title
        self.frame = frame
        self.separator = "   ::   "
        self.dirty_flag = False  # Flag to track if data was modified
        self.odd = False

        self.title_label = Widge.CLabel(self.frame, text=self.title, width=80, style='Info.TLabel')
        self.status = Widge.CLabel(self.frame, text="Highlight items above to remove and click Remove.", width=80, style='Info.TLabel')
        self.scrollbar = Scrollbar(self.frame)

        self.tree = ttk.Treeview(self.frame, style="Plain.Treeview")  # , selectmode="browse")
        self.tree.tag_configure('odd', background=AppStyle.ODD_ROW_COLOR)
        self.tree.tag_configure('even', background='white')

        self.tree["columns"] = ("pre",)
        self.tree.column("#0", width=400, minwidth=100, stretch=tk.NO)
        self.tree.column("pre", width=500, minwidth=50, stretch=tk.NO)
        self.tree.heading("#0", text="Name", anchor=tk.W)
        self.tree.heading("pre", text="  ", anchor=tk.W)

        self.tree.config(yscrollcommand=self.scrollbar.set)
        self.scrollbar.config(command=self.tree.yview)

        self.remove_button = ttk.Button(self.frame, text="remove", command=self.delete_handler, width=BUTTON_WIDTH)

        # Load in list from cache file
        self.directory = dir_name
        self.cache = CachedDictionary.CachedDictionary(dir_name, cache_filename)
        self.cache.read()
        self.dict = self.cache.dict
        self.logger.debug(f'{self.title}')

        # Configure buttons and widgets
        self.configure_widgets(frame)

        # Display data
        self.load_handler()

    def list_insert(self, tree, col1, col2):
        self.odd = not self.odd
        if self.odd:
            tag = odd_tag
        else:
            tag = even_tag
        tree.insert(parent='', index="end", iid=None, text=col1, values=(col2,), tags=tag)

    def clear_display_list(self, tree):
        self.odd = False
        for row in tree.get_children():
            tree.delete(row)

    def load_handler(self):
        # Load in list and display
        self.clear_display_list(self.tree)

        for item in sorted(self.dict):
            if len(self.dict[item]) > 1:
                self.list_insert(self.tree, f"{item}", f"{self.dict[item]}")
            else:
                self.list_insert(self.tree, f"{item}", '')

    def load_defaults(self):
        for item in default:
            self.dict[item] = ""  # Add item to dict
        self.add_handler()

    def delete_handler(self):
        # Delete selected items in list
        self.delete_items(self.tree, self.dict)
        self.dirty_flag = True

    def add_handler(self):
        # add  item to list
        self.load_handler()  # Reload listbox with new data
        self.dirty_flag = True

    """
    def get_list_selection(self):
        # Get the items the user selected in list (tree)
        col1 = (self.tree.item(self.tree.selection(), "text"))
        col2 = (self.tree.item(self.tree.selection())['values'][0])
        return f'{prefix}, {loc}'
    """

    def delete_items(self, tree, dct):
        # Delete any items in the list that the user selected
        items = tree.selection()
        for line in items:
            col1 = self.tree.item(line, "text")
            # col2 =
            self.logger.debug(f'DEL {col1}')
            dct.pop(col1, None)

        self.load_handler()  # Reload display

    def is_dirty(self):
        return self.dirty_flag  # Tells whether the cache was modified

    def configure_widgets(self, frm):
        Widge.TKHelper.set_grid_position(self.title_label, "title_label", grd=self.grd)
        Widge.TKHelper.set_grid_position(self.status, "status", grd=self.grd)
        Widge.TKHelper.set_grid_position(self.tree, "listbox", grd=self.grd)
        Widge.TKHelper.set_grid_position(self.scrollbar, "scrollbar", grd=self.grd)
        Widge.TKHelper.set_grid_position(self.remove_button, "remove_button", grd=self.grd)

    def write(self):
        # Write out cache file
        self.cache.write()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

import logging
import os
import time
import unittest
from pathlib import Path

from geofinder import Geodata, GeoKeys, Loc

halifax_lat = 44.71314


class TestGeodata(unittest.TestCase):
    geodata = None

    """
    Test case - multiple st andrews:
    6155128	St. Andrews	St. Andrews		46.38341	-62.84866	P	PPLL	CA		09			0
    6155129	St. Andrews	St. Andrews		45.55614	-61.88909	P	PPL	    CA		07			0
    6155132	St. Andrews	St. Andrews		50.0714	    -96.98393	P	PPLL	CA		03			0
    6137411	St. Andrews	St. Andrews		45.0737	    -67.05312	P	PPL	    CA		04	        1302
    """

    @classmethod
    def setUpClass(cls):
        logger = logging.getLogger(__name__)
        fmt = "%(levelname)s %(name)s.%(funcName)s %(lineno)d: %(message)s"
        logging.basicConfig(level=logging.DEBUG, format=fmt)

        # Load test data
        directory = os.path.join(str(Path.home()), "geoname_test")
        TestGeodata.geodata = Geodata.Geodata(directory_name=directory, progress_bar=None)
        error: bool = TestGeodata.geodata.read()
        if error:
            logger.error("Missing geodata support Files.")
            raise ValueError('Cannot open database')

        # Read in Geoname Gazeteer file - city names, lat/long, etc.
        start_time = time.time()
        error = TestGeodata.geodata.read_geonames()
        end_time = time.time()

        print(f'Elapsed {end_time - start_time}')
        if error:
            logger.info("Missing geoname Files.")
            logger.info('Requires ca.txt and de.txt from geonames.org in folder username/geoname_test')
            raise ValueError('Cannot open database')

    def setUp(self) -> None:
        self.place: Loc.Loc = Loc.Loc()

    def run_test(self, title: str, entry: str) -> float:
        print("*****TEST: {}".format(title))
        TestGeodata.geodata.find_location(entry, self.place)
        return float(self.place.lat)

    """
    Test all Result Types
        
    SUCCESS = 0 xx
    EOF = 1
    NOT_UNIQUE = 2  xx
    NOT_FOUND = 3  xx
    NOT_SUPPORTED = 4 xx
    NO_COUNTRY = 5  xx
    NOT_IDENTICAL = 7 xx
    """

    # ===== TEST RESULT CODES
    def test_res_code01(self):
        test = "City.  Good.  upper lowercase"
        lat: float = self.run_test(test, "AlbAnel,, Quebec, CanAda")
        self.assertEqual(GeoKeys.Result.EXACT_MATCH, self.place.result_type, test)

    def test_res_code02(self):
        test = "City - multiple matches"
        lat: float = self.run_test(test, "Alberton,, Ontario, Canada")
        self.assertEqual(GeoKeys.Result.MULTIPLE_MATCHES, self.place.result_type, test)

    def test_res_code03(self):
        test = "County - Good.  wrong Province"
        lat: float = self.run_test(test, "Halifax County, Alberta, Canada")
        self.assertEqual(GeoKeys.Result.PARTIAL_MATCH, self.place.result_type, test)

    def test_res_code11(self):
        test = "City and county  Good."
        lat: float = self.run_test(test, "baldwin mills,estrie,,canada")
        self.assertEqual(GeoKeys.Result.EXACT_MATCH, self.place.result_type, test)

    def test_res_code04(self):
        test = "city - Good. wrong Province"
        lat: float = self.run_test(test, "Halifax, ,Alberta, Canada")
        self.assertEqual(GeoKeys.Result.PARTIAL_MATCH, self.place.result_type, test)

    def test_res_code05(self):
        test = "multiple county - not unique"
        lat: float = self.run_test(test, "St Andrews,,,Canada")
        self.assertEqual(GeoKeys.Result.MULTIPLE_MATCHES, self.place.result_type, test)

    def test_res_code06(self):
        test = "City - good. wrong province"
        lat: float = self.run_test(test, "Natuashish, ,Alberta, Canada")
        self.assertEqual(GeoKeys.Result.PARTIAL_MATCH, self.place.result_type, test)

    def test_res_code07(self):
        test = "City - good. wrong county"
        lat: float = self.run_test(test, "Natuashish, Alberta, ,Canada")
        self.assertEqual(GeoKeys.Result.PARTIAL_MATCH, self.place.result_type, test)

    def test_res_code08(self):
        test = "City - Bad"
        lat: float = self.run_test(test, "Alberton, ,,Germany")
        self.assertEqual(GeoKeys.Result.NO_MATCH, self.place.result_type, test)

    def test_res_code09(self):
        test = "State - Bad"
        lat: float = self.run_test(test, "skdfjd,Germany")
        self.assertEqual(GeoKeys.Result.NO_MATCH, self.place.result_type, test)

    def test_res_code10(self):
        test = "Country - blank"
        lat: float = self.run_test(test, '')
        self.assertEqual(GeoKeys.Result.NO_COUNTRY, self.place.result_type, test)

    # Country
    def test_res_code_country01(self):
        test = "Country - bad"
        lat: float = self.run_test(test, "squid")
        self.assertEqual(GeoKeys.Result.MULTIPLE_MATCHES, self.place.result_type, test)

    def test_res_code_country02(self):
        test = "No Country - Natuashish"
        lat: float = self.run_test(test, "Natuashish,, ")
        self.assertEqual(GeoKeys.Result.NO_COUNTRY, self.place.result_type, test)

    def test_res_code_country03(self):
        test = "No Country - Berlin"
        lat: float = self.run_test(test, "Berlin,,, ")
        self.assertEqual(GeoKeys.Result.NO_COUNTRY, self.place.result_type, test)

    def test_res_code_country04(self):
        test = "Country - not supported"
        lat: float = self.run_test(test, "Tokyo,,,Japan")
        self.assertEqual(GeoKeys.Result.NOT_SUPPORTED, self.place.result_type, test)

    def test_res_code_country05(self):
        test = "Country - not supported"
        lat: float = self.run_test(test, "Tokyo,Japan")
        self.assertEqual(GeoKeys.Result.NOT_SUPPORTED, self.place.result_type, test)

    # =====  TEST PLACE TYPE CODES
    def test_place_code01(self):
        test = "Country  verify place type"
        lat: float = self.run_test(test, "Germany")
        self.assertEqual(Loc.PlaceType.COUNTRY, self.place.place_type, test)

    def test_place_code02(self):
        test = "State - Bad.  verify place type"
        lat: float = self.run_test(test, "skdfjd,Germany")
        self.assertEqual(Loc.PlaceType.ADMIN1, self.place.place_type, test)

    def test_place_code03(self):
        test = "State - Bad.  verify place type.  with prefix"
        lat: float = self.run_test(test, "abc,,,Alberta,Canada")
        self.assertEqual(Loc.PlaceType.ADMIN1, self.place.place_type, test)

    def test_place_code04(self):
        test = "County  prioritize city.  verify place type "
        lat: float = self.run_test(test, "Halifax, Nova Scotia, Canada")
        self.assertEqual(Loc.PlaceType.ADMIN2, self.place.place_type, test)

    def test_place_code24(self):
        test = "County  prioritize city.  verify place type "
        lat: float = self.run_test(test, "Halifax County, Nova Scotia, Canada")
        self.assertEqual(Loc.PlaceType.ADMIN2, self.place.place_type, test)

    def test_place_code05(self):
        test = "County prioritize city verify place type with prefix "
        lat: float = self.run_test(test, "abc,,Halifax, Nova Scotia, Canada")
        self.assertEqual(Loc.PlaceType.ADMIN2, self.place.place_type, test)

    def test_place_code25(self):
        test = "County prioritize city verify place type with prefix "
        lat: float = self.run_test(test, "abc,,Halifax County, Nova Scotia, Canada")
        self.assertEqual(Loc.PlaceType.ADMIN2, self.place.place_type, test)

    def test_place_code06(self):
        test = "City  verify place type"
        lat: float = self.run_test(test, "Halifax, , Nova Scotia, Canada")
        self.assertEqual(Loc.PlaceType.CITY, self.place.place_type, test)

    def test_place_code07(self):
        test = "City  verify place type with prefix"
        lat: float = self.run_test(test, "abc,,Halifax, , Nova Scotia, Canada")
        self.assertEqual(Loc.PlaceType.CITY, self.place.place_type, test)

    # ===== TEST PERMUTATIONS for Exact lookups (non wildcard)

    # Country -------------
    def test_country01(self):
        test = "Country -  good"
        lat: float = self.run_test(test, "Canada")
        self.assertEqual('canada', self.place.country_name, test)

    def test_country02(self):
        test = "Country -  bad"
        lat: float = self.run_test(test, "xyzzy")
        self.assertEqual('', self.place.country_iso, test)

    def test_country03(self):
        test = "Country - Good"
        lat: float = self.run_test(test, "Canada")
        self.assertEqual(60.0, lat, test)

    # Province -------------
    def test_province01(self):
        test = "Province - Good"
        lat: float = self.run_test(test, "Alberta, Canada")
        self.assertEqual(52.28333, lat, test)

    def test_province02(self):
        test = "Province - Good with prefix"
        lat: float = self.run_test(test, "abcde, , ,Alberta, Canada")
        self.assertEqual(52.28333, lat, test)

    def test_province03(self):
        test = "Province - Partial "
        lat: float = self.run_test(test, "new brunswick,Canada")
        self.assertEqual(46.5001, lat, test)

    def test_province04(self):
        test = "Province - bad name"
        lat: float = self.run_test(test, "xyzzy,Canada")
        self.assertEqual('', self.place.admin1_id, test)

    # County ---------------
    def test_county01(self):
        test = "County - good"
        lat: float = self.run_test(test, "Albert County, New Brunswick, Canada")
        self.assertEqual(45.83346, lat, test)

    def test_county02(self):
        test = "County - good with prefix"
        lat: float = self.run_test(test, "abcd, ,Albert County, New Brunswick, Canada")
        self.assertEqual(45.83346, lat, test)

    def test_county03(self):
        test = "County - Abbreviated good"
        lat: float = self.run_test(test, "Albert Co., New Brunswick, Canada")
        self.assertEqual(45.83346, lat, test)

    def test_county04(self):
        test = "County - good.  prioritize Halifax city vs County"
        lat: float = self.run_test(test, "Halifax, Nova Scotia, Canada")
        self.assertEqual(44.86685, lat, test)

    def test_county24(self):
        test = "County - good.  prioritize Halifax city vs County"
        lat: float = self.run_test(test, "Halifax County, Nova Scotia, Canada")
        self.assertEqual(44.86685, lat, test)

    # City -------------------
    def test_city01(self):
        test = "City - good. upper lowercase"
        lat: float = self.run_test(test, "AlbAnel,, Quebec, CanAda")
        self.assertEqual(48.88324, lat, test)

    def test_city02(self):
        test = "City - good, no county"
        lat: float = self.run_test(test, "Albanel,, Quebec, CanAda")
        self.assertEqual(48.88324, lat, test)

    def test_city03(self):
        test = "City - Good name, Saint"
        lat: float = self.run_test(test, "Saint Andrews,,Nova Scotia,Canada")
        self.assertEqual(45.55614, lat, test)

    def test_city04(self):
        test = "City - Good name, St "
        lat: float = self.run_test(test, "St Andrews,,Nova Scotia,Canada")
        self.assertEqual(45.55614, lat, test)

    def test_city05(self):
        test = "City - Good name, edberg "
        lat: float = self.run_test(test, "Edberg,,Alberta,Canada")
        self.assertEqual(52.78343, lat, test)

    def test_city07(self):
        test = "City - Good name, gray gull island "
        lat: float = self.run_test(test, "gray gull island,,newfoundland and labrador,Canada")
        self.assertEqual(47.5166, lat, test)

    def test_city08(self):
        test = "City - Good name, grey gull island with E "
        lat: float = self.run_test(test, "grey gull island,,newfoundland and labrador,Canada")
        self.assertEqual(47.5166, lat, test)

    def test_city09(self):
        test = "City - Good name, St. "
        lat: float = self.run_test(test, "St. Andrews,,Nova Scotia,Canada")
        self.assertEqual(45.55614, lat, test)

    def test_city10(self):
        test = "City - Good name, St. with county and province"
        lat: float = self.run_test(test, "St. Andrews,Charlotte County,new brunswick,Canada")
        self.assertEqual(45.0737, lat, test)

    def test_city11(self):
        test = "City - Good , Alberton PEI vs Alberton Ontario"
        lat: float = self.run_test(test, "Alberton,,Prince Edward Island, Canada")
        self.assertEqual(46.81685, lat, test)

    def test_city12(self):
        test = "City - Good , Alberton Ontario vs Alberton PEI"
        lat: float = self.run_test(test, "Alberton,Rainy River District,Ontario, Canada")
        self.assertEqual(48.58318, lat, test)

    def test_city13(self):
        test = "City - Good , Halifax, Nova Scotia"
        lat: float = self.run_test(test, "Halifax,, Nova Scotia, Canada")
        self.assertEqual(halifax_lat, lat, test)

    def test_city14(self):
        test = "City - Good - with prefix"
        lat: float = self.run_test(test, "Oak Street, Halifax, ,Nova Scotia, Canada")
        self.assertEqual(halifax_lat, lat, test)

    def test_city15(self):
        test = "City - Good - with prefix and wrong county"
        lat: float = self.run_test(test, "Oak Street, Halifax, aaa, Nova Scotia, Canada")
        self.assertEqual(halifax_lat, lat, test)

    def test_city16(self):
        test = "City -  Good - no county"
        lat: float = self.run_test(test, "St Andrews,,Nova Scotia,Canada")
        self.assertEqual(45.55614, lat, test)

    def test_city17(self):
        test = "County - no County"
        lat: float = self.run_test(test, "Albanel, ,Quebec, Canada")
        self.assertEqual(48.88324, lat, test)

    def test_city18(self):
        test = "County - Wrong county"
        lat: float = self.run_test(test, "Albanel, Zxzzy, Quebec, Canada")
        self.assertEqual(48.88324, lat, test)

    def test_city19(self):
        test = "City - good - Wrong Province"
        lat: float = self.run_test(test, "Halifax, Alberta, Canada")
        self.assertAlmostEqual(-63.71541, float(self.place.lon))

    def test_city20(self):
        test = "City - Good "
        lat: float = self.run_test(test, "Natuashish,,, Canada")
        self.assertEqual(55.91564, lat, test)

    def test_city21(self):
        test = "City - Good "
        lat: float = self.run_test(test, "Natuashish")
        self.assertEqual(55.91564, lat, test)

    def test_city22(self):
        test = "City -  wrong county but single match"
        lat: float = self.run_test(test, "Agassiz,, british columbia, Canada")
        self.assertEqual(49.23298, lat, test)

    def test_city23(self):
        test = "City - good. No Country"
        lat: float = self.run_test(test, "Albanel,,Quebec")
        self.assertEqual(48.88324, lat, test)

    def test_city24(self):
        test = "City - good. city,country"
        lat: float = self.run_test(test, "Natuashish,canada")
        self.assertEqual(55.91564, lat, test)

    def test_city25(self):
        test = "City - from AlternateNames"
        lat: float = self.run_test(test, "Pic du port, canada")
        self.assertEqual(52.28333, lat, test)

    # ===== TEST WILDCARDS
    def test_wildcard01(self):
        test = "Country - wildcard country"
        lat: float = self.run_test(test, "C*da")
        self.assertEqual('canada', self.place.country_name, test)

    def test_wildcard02(self):
        test = "Province - wildcard province"
        lat: float = self.run_test(test, "Al*ta, Canada")
        self.assertEqual(52.28333, lat, test)

    def test_wildcard03(self):
        test = "City - good - wildcard city, full county and province"
        lat: float = self.run_test(test, "St. Andr,Charlotte County,new brunswick,Canada")
        self.assertEqual(45.0737, lat, test)

    def test_wildcard04(self):
        test = "City - good - wildcard city, no county"
        lat: float = self.run_test(test, "Alb*el,, Quebec, CanAda")
        self.assertEqual(48.88324, lat, test)

    def test_wildcard05(self):
        test = "City - good - wildcard city, no county, wild province"
        lat: float = self.run_test(test, "Alb*el,, Queb*, CanAda")
        self.assertEqual(48.88324, lat, test)

    def test_wildcard06(self):
        test = "City - good - wildcard city, no county, wild province, wild country"
        lat: float = self.run_test(test, "Alb*el,, Queb*, Can*a")
        self.assertEqual(48.88324, lat, test)

    def test_wildcard07(self):
        test = "City - Wildcard - 300+ matches"
        lat: float = self.run_test(test, "b*,,, CanAda")
        self.assertEqual(GeoKeys.Result.MULTIPLE_MATCHES, self.place.result_type, test)

    # ===== TEST ADMIN ID

    def test_admin_id01(self):
        test = "Admin2 ID - good "
        lat: float = self.run_test(test, "Alberton,Rainy River District,Ontario, Canada")
        self.assertEqual('3559', self.place.admin2_id, test)

    def test_admin_id02(self):
        test = "Admin2 ID - good, no province "
        lat: float = self.run_test(test, "Alberton,Rainy River District, , Canada")
        self.assertEqual('3559', self.place.admin2_id, test)

    def test_admin_id03(self):
        test = "Admin1 ID - good "
        lat: float = self.run_test(test, "Nova Scotia,Canada")
        self.assertEqual('07', self.place.admin1_id, test)

    def test_admin_id04(self):
        test = "Admin1 ID - good - abbreviated "
        lat: float = self.run_test(test, "Baden, Germany")
        self.assertEqual('01', self.place.admin1_id, test)

    def test_admin_id05(self):
        test = "Admin1 ID - good.  With non-ASCII"
        lat: float = self.run_test(test, "Baden-Württemberg Region, Germany")
        self.assertEqual('01', self.place.admin1_id, test)

    def test_admin_id06(self):
        test = "Admin1 ID - good - abbreviated "
        lat: float = self.run_test(test, "Nova,Canada")
        self.assertEqual('07', self.place.admin1_id, test)

    def test_admin_id07(self):
        test = "Admin1 ID - good - abbreviated, non-ASCII "
        lat: float = self.run_test(test, "Baden-Württemberg, Germany")
        self.assertEqual('01', self.place.admin1_id, test)

    # ===== TEST PARSING
    def test_parse01(self):
        test = "***** Test Parse admin1"
        print(test)
        self.place.parse_place(place_name="aaa,Banff,Alberta's Rockies,Alberta,Canada", geo_files=TestGeodata.geodata.geo_files)
        self.assertEqual("alberta", self.place.admin1_name, test)

    def test_parse02(self):
        test = "***** Test Parse admin2"
        print(test)
        self.place.parse_place(place_name="aaa,Banff,Alberta's Rockies,Alberta,Canada", geo_files=TestGeodata.geodata.geo_files)
        self.assertEqual("alberta's rockies", self.place.admin2_name, test)

    def test_parse03(self):
        test = "***** Test Parse city with punctuation"
        print(test)
        self.place.parse_place(place_name="aaa,Banff!@#^&)_+-=;:<>/?,Alberta's Rockies,Alberta,Canada", geo_files=TestGeodata.geodata.geo_files)
        self.assertEqual("banff", self.place.city1, test)

    def test_parse04(self):
        test = "***** Test Parse prefix"
        print(test)
        self.place.parse_place(place_name="pref,   abcde,Banff,Alberta's Rockies,Alberta,Canada", geo_files=TestGeodata.geodata.geo_files)
        self.assertEqual("pref abcde", self.place.prefix + self.place.prefix_commas, test)

    # =====  TEST Name formatting
    def test_place_name01(self):
        test = "Country  verify place name"
        lat: float = self.run_test(test, "Germany")
        nm = self.place.format_full_name()
        self.assertEqual("Germany", self.place.prefix + self.place.prefix_commas + nm, test)

    def test_place_name03(self):
        test = "State - Bad.  verify place name.  with prefix"
        lat: float = self.run_test(test, "abc,,,Alberta,Canada")
        nm = self.place.format_full_name()
        self.assertEqual("abc, , ,  Alberta, Canada", self.place.prefix + self.place.prefix_commas + nm, test)

    def test_place_name04(self):
        test = "County  verify place name "
        lat: float = self.run_test(test, "Halifax, Nova Scotia, Canada")
        nm = self.place.format_full_name()
        self.assertEqual("Halifax County, Nova Scotia, Canada", self.place.prefix + self.place.prefix_commas + nm, test)

    def test_place_name05(self):
        test = "County  verify place name with prefix. prioritize city "
        lat: float = self.run_test(test, "abc,,Halifax, Nova Scotia, Canada")
        nm = self.place.format_full_name()
        self.assertEqual("abc, , Halifax County, Nova Scotia, Canada", self.place.prefix + self.place.prefix_commas + nm, test)

    def test_place_name25(self):
        test = "County  verify place name with prefix "
        lat: float = self.run_test(test, "abc,,Halifax County, Nova Scotia, Canada")
        nm = self.place.format_full_name()
        self.assertEqual("abc, , Halifax County, Nova Scotia, Canada", self.place.prefix + self.place.prefix_commas + nm, test)

    def test_place_name06(self):
        test = "City  verify place name"
        lat: float = self.run_test(test, "Halifax, , Nova Scotia, Canada")
        nm = self.place.format_full_name()
        self.assertEqual("Halifax, , Nova Scotia, Canada", self.place.prefix + self.place.prefix_commas + nm, test)

    def test_place_name07(self):
        test = "City  verify place name with prefix"
        lat: float = self.run_test(test, "abc,,Halifax, , Nova Scotia, Canada")
        nm = self.place.format_full_name()
        self.assertEqual("abc , Halifax, , Nova Scotia, Canada", self.place.prefix + self.place.prefix_commas + nm, test)

    def test_place_name08(self):
        test = "province  verify place name with country"
        lat: float = self.run_test(test, "Alberta")
        nm = self.place.format_full_name()
        self.assertEqual(" Alberta, Canada", self.place.prefix + self.place.prefix_commas + nm, test)

    # ======= TEST Event Year
    def test_eventyear01(self):
        test = "City - good - and after city start"
        self.place.event_year = 1541
        lat: float = self.run_test(test, "Albanel,, Quebec, Canada")
        self.assertEqual(48.88324, lat, test)

    def test_eventyear02(self):
        test = "City - good - but before city start"
        self.place.event_year = 1540
        lat: float = self.run_test(test, "Albanel,, Quebec, Canada")
        self.assertEqual(GeoKeys.Result.PARTIAL_MATCH, self.place.result_type, test)

    def test_eventyear03(self):
        test = "City - not in list"
        self.place.event_year = 1541
        lat: float = self.run_test(test, "Stuttgart,,,Germany")
        self.assertEqual(48.78232, lat, test)

    # ====== TEST find first match, find geoid
    def test_findfirst01(self):
        test = "City - find first"
        entry = "Berlin,,,Germany"

        print("*****TEST: {}".format(test))
        TestGeodata.geodata.find_first_match(entry, self.place)
        lat = float(self.place.lat)

        self.assertEqual(54.03573, lat, test)

    def test_findgeoid01(self):
        test = "City - find first"
        entry = "Berlin,,,Germany"

        print("*****TEST: {}".format(test))
        TestGeodata.geodata.find_geoid('8133394', self.place)
        lat = float(self.place.lat)

        self.assertEqual(43.69655, lat, test)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

# The tab separated columns in geoname.org file rows are as follows

from geofinder import GeodataFiles, GeoKeys, GeoDB
from geofinder.FileReader import FileReader
from geofinder.Loc import Loc

ALT_GEOID = 1
ALT_LANG = 2
ALT_NAME = 3


class AlternateNames(FileReader):
    """
    Read in alternate names file and add appropriate entries to geoname dictionary
    Each row contains a geoname ID, an alternative name for that entity, and the language
    If the lang is english and the ID is in our geonames dictionary, we add this as an alternative name
    FileReader calls handle_line every time it reads a line
    """

    def __init__(self, directory_name: str, filename: str, progress_bar, geo_files: GeodataFiles, lang_list):
        super().__init__(directory_name, filename, progress_bar)
        self.sub_dir = GeoKeys.get_cache_directory(directory_name)
        self.geo_files: GeodataFiles.GeodataFiles = geo_files
        self.lang_list = lang_list
        self.loc = Loc()

    def read(self) -> bool:
        self.geo_files.geodb.db.begin()
        # Read in file.  This will call handle_line for each file line
        res = super().read()
        self.geo_files.geodb.db.commit()
        return res

    def cancel(self):
        # User requested cancel
        # Abort DB build.  Clear out partial DB
        self.geo_files.geodb.db.commit()
        self.geo_files.geodb.clear_geoname_data()

    def handle_line(self, line_num, row):
        # This is called as each line is read
        alt_tokens = row.split('\t')
        if len(alt_tokens) != 10:
            self.logger.debug(f'Incorrect number of tokens: {alt_tokens} line {line_num}')
            return

        self.loc.georow_list = []

        # Alternate names are in multiple languages.  Only add if item is in requested lang
        if alt_tokens[ALT_LANG] in self.lang_list:
            # Add this alias to geoname db if there is already an entry (geoname DB is filtered based on feature)
            # See if item has a primary entry with same GEOID in Main DB
            dbid = self.geo_files.geodb.geoid_main_dict.get(alt_tokens[ALT_GEOID])
            if dbid is not None:
                self.loc.target = dbid
                self.geo_files.geodb.lookup_main_dbid(place=self.loc)
            else:
                # See if item has a primary entry with same GEOID in Admin DB
                dbid = self.geo_files.geodb.geoid_admin_dict.get(alt_tokens[ALT_GEOID])
                if dbid is not None:
                    self.loc.target = dbid
                    self.geo_files.geodb.lookup_admin_dbid(place=self.loc)

            if len(self.loc.georow_list) > 0:
                # convert to list  and modify name and add to DB and its soundex
                lst = list(self.loc.georow_list[0])
                lst[GeoDB.Entry.NAME] = GeoKeys.normalize(alt_tokens[ALT_NAME])
                lst.append(GeoKeys.get_soundex(alt_tokens[ALT_NAME]))
                new_row = tuple(lst)   # Convert back to tuple

                self.geo_files.geodb.insert(geo_row=new_row, feat_code=lst[GeoDB.Entry.FEAT])
                self.count += 1



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
import logging
import os
import webbrowser
from tkinter import *
from tkinter import ttk
from tkinter.ttk import *
from typing import List

from geofinder import UtilCountriesFrame, UtilErrorFrame, UtilFeatureFrame, UtilReplaceFrame, ListboxFrame, AppStyle, UtilLanguagesFrame


class UtilLayout:
    """ Create Configuration utility windows"""
    def __init__(self, root, directory, cache_dir):
        self.logger = logging.getLogger(__name__)
        self.directory = directory
        self.cache_dir = cache_dir
        self.root = root

    def start_up(self):
        # Setup styles
        self.root.configure(background=AppStyle.BG_COLOR)
        AppStyle.GFStyle()
        self.create_util_widgets()

        # Start up
        self.root.mainloop()  # Loop getting user actions. We are now controlled by user window actions

    def create_util_widgets(self):
        self.root.title("GeoFinder Configuration Utility")
        self.frames: List[Frame] = []
        self.listbox_list: List[ListboxFrame] = []
        self.error = ""

        # Add Multiple tabs
        tab_list = ["Errors", "Countries", "Skip List", "Global Replace", "Features", "Languages"]
        self.create_tabs(tab_list)

        # Create a frame for each tab:

        # Error Status Tab
        self.logger.debug('=====error frame')
        self.status_list = UtilErrorFrame.SetupErrorFrame(self.frames[0], "Configuration Status",
                                                          self.directory, "errors.pkl", self.error)
        self.listbox_list.append(self.status_list)

        # Country Tab -  (has text entry box to add items)
        self.logger.debug('=====country frame')
        self.country_list = UtilCountriesFrame.SetupCountriesFrame(self.frames[1],
                                                                    "Supported Countries - Load geo data for these countries:",
                                                                   self.directory, self.cache_dir, "country_list.pkl")
        self.listbox_list.append(self.country_list)

        # Skiplist Tab - ListboxFrame (simple list)
        self.logger.debug('=====skiplist frame')
        self.skip_list = ListboxFrame.ListboxFrame(self.frames[2], "Skiplist - Ignore errors for these places",
                                                   self.cache_dir, "skiplist.pkl")
        self.listbox_list.append(self.skip_list)

        # GlobalReplace Tab- ListboxFrame (simple list)
        self.logger.debug('=====gbl replace frame')
        self.replace_list = UtilReplaceFrame.SetupReplaceFrame(self.frames[3], "Global Replace  - Replace these errors",
                                                               self.cache_dir, "global_replace.pkl")
        self.listbox_list.append(self.replace_list)

        # Feature tab
        self.logger.debug('=====Feature frame')
        self.feature_list = UtilFeatureFrame.SetupFeatureFrame(self.frames[4],
                                                                "We will load data for these geoname feature types:",
                                                               self.cache_dir, "feature_list.pkl")
        self.listbox_list.append(self.feature_list)

        # Languages tab
        self.logger.debug('=====Languages frame')
        self.languages_list = UtilLanguagesFrame.UtilLanguagesFrame(self.frames[5],
                                                                "Load alternate names for these languages:",
                                                               self.cache_dir, "languages_list.pkl")
        self.listbox_list.append(self.languages_list)

        # Create Help button below frames
        self.help_button = ttk.Button(self.root, text="help", command=self.help_handler, width=10)
        self.help_button.grid(row=1, column=0, sticky="E", pady=9, padx=8)

        # Create Quit button below frames
        self.quit_button = ttk.Button(self.root, text="quit", command=self.quit_handler, width=10)
        self.quit_button.grid(row=1, column=1, sticky="E", pady=9, padx=8)

        #  write out all cache files
        for item in self.listbox_list:
            item.write()

    def create_tabs(self, tabs_list):
        nb = ttk.Notebook(self.root, style='TNotebook')
        nb.grid(row=0, column=0, columnspan=2, pady=9, padx=8)
        for element in tabs_list:
            frame = Frame()
            nb.add(frame, text=element)
            self.frames.append(frame)

    @staticmethod
    def help_handler():
        """ Bring up browser with help text """
        help_base = "https://github.com/corb555/GeoFinder/wiki/User-Guide"
        webbrowser.open(help_base)

    def quit_handler(self):
        #  write out all cache files
        for item in self.listbox_list:
            item.write()

        if self.country_list.is_dirty() or self.feature_list.is_dirty() or self.languages_list.is_dirty():
            # Delete geoname.db so GeoFinder will rebuild it with new country list or feature list
            for fname in ['geodata.db']:
                path = os.path.join(self.cache_dir, fname)
                self.logger.debug(f'Quit - DELETING FILE {path}')
                if os.path.exists(path):
                    os.remove(path)
                else:
                    self.logger.warning(f'Delete file not found {path}')

        self.root.quit()
        sys.exit()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
import collections
import os
import re

import phonetics
import unidecode


class Entry:
    NAME = 0
    ISO = 1
    ADM1 = 2
    ADM2 = 3
    LAT = 4
    LON = 5
    FEAT = 6
    ID = 7
    SDX = 8
    MAX = 9


class Result:
    # Result codes for lookup
    EXACT_MATCH = 0
    MULTIPLE_MATCHES = 2
    NO_MATCH = 3
    NOT_SUPPORTED = 4
    NO_COUNTRY = 5
    PARTIAL_MATCH = 6
    DELETE = 7


# Result types that are successful matches
successful_match = [Result.EXACT_MATCH, Result.PARTIAL_MATCH]

Query = collections.namedtuple('Query', 'where args result')


def get_soundex(txt):
    res = phonetics.dmetaphone(txt)
    return res[0]


def get_directory_name() -> str:
    return "geoname_data"


def get_cache_directory(dirname):
    """ Return the directory for cache files """
    return os.path.join(dirname, "cache")


def _phrase_normalize(name) -> str:
    """ Strip spaces and normalize spelling for items such as Saint and County """
    # Replacement patterns to clean up entries
    res = re.sub('saints |sainte |sint |saint |sankt |st. ', 'st ', name)  # Normalize Saint
    res = re.sub(r' co\.', ' county', res)  # Normalize County
    res = re.sub('  +', ' ', res)  # Strip multiple space
    res = re.sub('county of ([^,]+)', r'\g<1> county', res)  # Normalize 'Township of X' to 'X Township'
    res = re.sub('township of ([^,]+)', r'\g<1> township', res)  # Normalize 'Township of X' to 'X Township'
    res = re.sub('cathedral of ([^,]+)', r'\g<1> cathedral', res)  # Normalize 'Township of X' to 'X Township'
    return res


def normalize(res) -> str:
    """ Strip commas. Also strip spaces and normalize spelling for items such as Saint and County and chars   ø ß """

    # Convert UT8 to ascii
    res = unidecode.unidecode(res)

    res = str(res).lower()

    # remove all punctuation
    res = re.sub(r"[^a-zA-Z0-9 $.*']+", " ", res)

    res = _phrase_normalize(res)
    return res.strip(' ')


type_names = {
    "CH": 'Church',
    "CSTL": 'Castle',
    "CMTY": 'Cemetery',
    "EST": 'Estate',
    "HSP": 'Hospital',
    "HSTS": 'Historical',
    "ISL": 'Island',
    "MT": 'Mountain',
    "MUS": 'Museum',
    "PAL": 'Palace',
    "PRK": 'Park',
    "PRN": 'Prison',
    "RUIN": 'Ruin',
    "SQR": 'Square',
    "VAL": 'Valley',
    "ADM1": 'State',
    "ADM2": 'County',
    "ADM3": 'Township',
    "ADM4": 'Township',
    "PPL": 'City',
    "PPLA": 'City',
    "PPLA2": 'City',
    "PPLA3": 'City',
    "PPLA4": 'City',
    "PPLC": 'City',
    "PPLG": 'City',
    "PPLH": 'City',
    "PPLL": 'Village',
    "PPLQ": 'Historical',
    "PPLX": 'Neighborhood'
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

import logging
import re
import string as st
from typing import List, Tuple

from geofinder import GeoKeys

default_country = 'nederland'


class Loc:
    """
    Holds the details about a Location: Name, county, state/province, country, lat/long as well as lookup result details
    Parses a name into Loc items (county, state, etc)
    """

    def __init__(self):
        self.logger = logging.getLogger(__name__)
        self.clear()
        self.event_year: int = 0

    def clear(self):
        # Place geo info
        self.name: str = ""
        self.lat: float = float('NaN')  # Latitude
        self.lon: float = float('NaN')  # Longitude
        self.country_iso: str = ""  # Country ISO code
        self.country_name: str = ''
        self.city1: str = ""  # City or entity name
        self.admin1_name: str = ""  # Admin1 (State/province/etc)
        self.admin2_name: str = ""  # Admin2 (county)
        self.admin1_id: str = ""  # Admin1 Geoname ID
        self.admin2_id = ""  # Admin2 Geoname ID
        self.prefix: str = ""  # Prefix (entries before city)
        self.feature: str = ''  # Geoname feature code
        self.place_type: int = PlaceType.COUNTRY  # Is this a Country , Admin1 ,admin2 or city?
        self.target: str = ''  # Target for lookup
        self.geoid: str = ''
        self.prefix_commas: str = ''

        # Lookup result info
        self.status: str = ""
        self.status_detail: str = ""
        self.result_type: int = GeoKeys.Result.NO_MATCH  # Result type of lookup
        self.result_type_text: str = ''  # Text version of result type
        self.georow_list: List[Tuple] = [()]  # List of items that matched this location

        self.georow_list.clear()

    def parse_place(self, place_name: str, geo_files):
        """
        Given a comma separated place name, parse into its city, AdminID, country_iso and type of entity (city, country etc)
        Expected format: prefix,city,admin2,admin1,country
        Place.status has Result status code
        """
        self.clear()
        self.name = place_name

        # Convert open-brace and open-paren to comma.  close brace/paren will be stripped by normalize()
        res = re.sub('\[', ',', place_name)
        res = re.sub('\(', ',', res)

        tokens = res.split(",")
        token_count = len(tokens)
        self.place_type = PlaceType.CITY

        # Parse City, Admin2, Admin2, Country scanning from the right.  When there are more tokens, we capture more fields
        # Place type is the leftmost item we found - either City, Admin2, Admin2, or Country

        #  COUNTRY - right-most token should be country
        if token_count > 0:
            #  Format: Country
            self.place_type = PlaceType.COUNTRY
            self.country_name = GeoKeys.normalize(tokens[-1])

            # Validate country
            self.country_iso = geo_files.geodb.get_country_iso(self)  # Get Country country_iso
            self.logger.debug(f'iso = [{self.country_iso}]')

            if self.country_iso is '':
                # Country not found.
                # See if rightmost token is actually Admin1 (state/province).  Look up admin1 and derive its country
                save_admin1 = self.admin1_name
                self.admin1_name = GeoKeys.normalize(self.country_name)

                # Lookup last token as admin1.  This will fill in country ISO if admin1 is found
                geo_files.geodb.get_admin1_id(self)
                self.admin1_name = save_admin1  # Restore Admin1

                if self.country_iso != '':
                    # We found the country.  Append it to token list so we now have xx,admin1, country
                    self.logger.debug(f'Last tkn is state/province {self.admin1_name}')
                    self.logger.debug(f'iso=[{self.country_iso}]')
                    tokens.append(geo_files.geodb.get_country_name(self.country_iso))
                    self.country_name = tokens[-1].strip(' ').lower()
                    token_count = len(tokens)
                else:
                    # No country found for this
                    # Append blank to token list so we now have xx,admin1, blank_country
                    self.logger.debug(f'didnt find last tkn as adm1 {self.admin1_name}')
                    tokens.append('')

                    token_count = len(tokens)
                    self.result_type = GeoKeys.Result.NO_COUNTRY
                    self.country_iso = ''

                    # Validate country
                    self.country_iso = geo_files.geodb.get_country_iso(self)  # Get Country country_iso

            self.target = self.country_name

        if token_count > 1:
            #  Format: Admin1, Country.
            #  Admin1 is 2nd to last token
            self.admin1_name = GeoKeys.normalize(tokens[-2])
            if len(self.admin1_name) > 0:
                self.place_type = PlaceType.ADMIN1
                self.target = self.admin1_name

        if token_count > 2:
            #  Format: Admin2, Admin1, Country
            #  Admin2 is 3rd to last.  Note -  if Admin2 isnt found, it will look it up as city
            self.admin2_name = GeoKeys.normalize(tokens[-3])
            if len(self.admin2_name) > 0:
                self.place_type = PlaceType.ADMIN2
                #if self.admin2_name == 'london':
                    # Change to Greater London
                    #self.admin2_name = 'greater london'
                self.target = self.admin2_name
            # else:
            #    self.place_type = PlaceType.CITY

        if token_count > 3:
            # Format: Prefix, City, Admin2, Admin1, Country
            # City is 4th to last token
            # Other tokens go into Prefix
            self.city1 = GeoKeys.normalize(tokens[-4])
            if len(self.city1) > 0:
                self.place_type = PlaceType.CITY
                self.target = self.city1

            # Assign remaining tokens (if any) to prefix
            for item in tokens[0:-4]:
                if len(self.prefix) > 0:
                    self.prefix += ' '
                self.prefix += item.strip(' ')

            # self.prefix.strip(' ')

        # Special case for New York, New York which normally refers to the City, not county
        if self.admin2_name == 'new york' and self.place_type == PlaceType.ADMIN2:
            self.admin2_name = 'new york city'
            self.target = self.admin2_name

        self.logger.debug(f"*** PARSE {token_count} tokens. City [{self.city1}] Adm2 [{self.admin2_name}]"
                          f"  Adm1 [{self.admin1_name}] Cntry [{self.country_name}] Pref=[{self.prefix}] Commas=[{self.prefix_commas}]"
                          f" Typ=[{place_type_name_dict[self.place_type]}] type_id={self.place_type}")
        return

    def set_place_type(self):
        self.place_type = PlaceType.CITY
        if len(str(self.country_name)) > 0:
            self.place_type = PlaceType.COUNTRY
        if len(self.admin1_name) > 0:
            self.place_type = PlaceType.ADMIN1
        if len(self.admin2_name) > 0:
            self.place_type = PlaceType.ADMIN2
        if len(self.city1) > 0:
            self.place_type = PlaceType.CITY

    def get_status(self) -> str:
        return self.status

    def format_full_name(self):
        """ Take the parts of a Place and build fullname.  e.g. pref, city,adm2,adm1,country name """
        self.set_place_type()
        self.prefix = self.prefix.strip(',')

        if self.admin1_name is None:
            self.admin1_name = ''
        if self.admin2_name is None:
            self.admin2_name = ''

        if self.place_type == PlaceType.COUNTRY:
            nm = f"{st.capwords(self.country_name)}"
            self.prefix_commas = ', , , , '
        elif self.place_type == PlaceType.ADMIN1:
            nm = f" {st.capwords(self.admin1_name)}, {st.capwords(self.country_name)}"
            self.prefix_commas = ', , , '
        elif self.place_type == PlaceType.ADMIN2:
            nm = f"{st.capwords(self.admin2_name)}," \
                f" {st.capwords(self.admin1_name)}, {st.capwords(self.country_name)}"
            self.prefix_commas = ', , '
        else:
            nm = f"{st.capwords(self.city1)}, {st.capwords(self.admin2_name)}, " \
                f"{st.capwords(self.admin1_name)}, {st.capwords(str(self.country_name))}"
            self.prefix_commas = ', '

        if len(self.prefix) == 0:
            self.prefix_commas = ''
        # self.logger.debug(f' [{self.prefix}][{self.prefix_commas}][{nm}]')
        return nm


# What type of entity is this place?
class PlaceType:
    COUNTRY = 0
    ADMIN1 = 1
    ADMIN2 = 2
    CITY = 3


place_type_name_dict = {
    PlaceType.COUNTRY: 'Country',
    PlaceType.ADMIN1: 'STATE/PROVINCE',
    PlaceType.ADMIN2: 'COUNTY',
    PlaceType.CITY: ' '
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
import logging
import tkinter as tk
from tkinter import ttk
from typing import List

import pkg_resources

import geofinder.AppStyle as GFStyle
import geofinder.TKHelper
from geofinder import Progress, Tooltip

# Columns for widgets:
# [0 PADDING]  [1 TEXT    with span 2]  [3 Buttons]
#                       [2 Scroll Bar]
PAD_COL = 0
TXT_COL = 1
SCRL_COL = 2
BTN_COL = 3


class AppLayout:

    def __init__(self, main):
        """ Create the app window and styles """
        self.logger = logging.getLogger(__name__)
        self.main = main

        self.root = tk.Tk()
        self.root.title("GeoFinder")
        self.root["padx"] = 0
        self.root["pady"] = 20

        # Set column/row weight for responsive resizing
        self.root.columnconfigure(0, weight=1)
        for rw in range(0, 12):
            self.root.rowconfigure(rw, weight=1)

        # Setup styles
        self.root.configure(background=GFStyle.BG_COLOR)
        GFStyle.GFStyle()

        # Load images for buttons
        self.images = {}
        for icon_name in ("map", "verify", "save", "skip", "help", "exit", "play", "folder", "search"):
            path = pkg_resources.resource_filename(__name__, f'images/{icon_name}.gif')
            self.images[icon_name] = tk.PhotoImage(file=path)

        self.root.update()

    def create_initialization_widgets(self):
        """ Create the  widgets for display during initialization  (File open)  """
        self.pad: geofinder.TKHelper.CLabel = geofinder.TKHelper.CLabel(self.root, text=" ", width=2, style='Light.TLabel')
        self.title: geofinder.TKHelper.CLabel = geofinder.TKHelper.CLabel(self.root, text="GEO FINDER", width=30, style='Large.TLabel')
        self.original_entry: geofinder.TKHelper.CLabel = geofinder.TKHelper.CLabel(self.root, text=" ", width=50, style='Info.TLabel')
        self.status: geofinder.TKHelper.CLabel = geofinder.TKHelper.CLabel(self.root, width=GFStyle.TXT_WID, style='Good.TLabel')
        self.prog: Progress.Progress = Progress.Progress(self.root, bar_color=GFStyle.HIGH_COLOR, trough_color=GFStyle.LT_GRAY, status=self.status)
        self.quit_button: ttk.Button = ttk.Button(self.root, text="quit", command=self.main.shutdown,
                                                  width=GFStyle.BTN_WID_WD, image=self.images['exit'], compound="left")
        self.load_button: ttk.Button = ttk.Button(self.root, text="open", command=self.main.load_handler,
                                                  width=GFStyle.BTN_WID_WD, style='Preferred.TButton', image=self.images['play'], compound="left")
        self.choose_button: ttk.Button = ttk.Button(self.root, text="choose", command=self.main.filename_handler,
                                                    width=GFStyle.BTN_WID_WD, image=self.images['folder'], compound="left")
        self.config_button: ttk.Button = ttk.Button(self.root, text="config", command=self.main.config_handler,
                                                    width=GFStyle.BTN_WID_WD, image=self.images['play'], compound="left")

        # Set grid layout for padding column widget - just pads out left column
        self.pad.grid(column=PAD_COL, row=0, padx=GFStyle.PAD_PADX, pady=0, sticky="EW")

        # Set grid for text widgets
        self.original_entry.grid(column=TXT_COL, row=2, padx=7, pady=5, sticky="EW")
        self.status.grid(column=TXT_COL, row=4, padx=7, pady=5, sticky="EW")
        self.title.grid(column=TXT_COL, row=0, padx=7, pady=12, sticky="")
        self.prog.bar.grid(column=TXT_COL, row=1, padx=7, pady=5, sticky="EW")

        # Set grid for button widgets
        self.load_button.grid(column=SCRL_COL, row=2, padx=20, pady=6, sticky="")
        self.choose_button.grid(column=SCRL_COL, row=3, padx=20, pady=6, sticky="")
        self.config_button.grid(column=SCRL_COL, row=4, padx=20, pady=6, sticky="")
        self.quit_button.grid(column=SCRL_COL, row=9, padx=20, pady=6, sticky="S")

        Tooltip.Tooltip(self.root, self.load_button, text="Open GEDCOM file")
        Tooltip.Tooltip(self.root, self.choose_button, text="Choose GEDCOM file")

        self.root.update()

        self.initialization_buttons: List[ttk.Button] = [self.quit_button, self.load_button, self.choose_button, self.config_button]
        # disable all buttons and the app will enable appropriate ones
        geofinder.TKHelper.TKHelper.disable_buttons(button_list=self.initialization_buttons)

    def remove_initialization_widgets(self):
        #self.w.original_entry.set_text("")
        self.original_entry.destroy()
        self.title.destroy()
        self.prog.bar.destroy()
        self.status.destroy()

        self.load_button.destroy()
        self.choose_button.destroy()
        self.quit_button.destroy()
        self.config_button.destroy()

    def create_review_widgets(self):
        """ Create all the buttons and entry fields for normal running """
        self.root.protocol("WM_DELETE_WINDOW", self.main.quit_handler)

        self.pad: geofinder.TKHelper.CLabel = geofinder.TKHelper.CLabel(self.root, text=" ", width=2, style='Light.TLabel')
        self.title: geofinder.TKHelper.CLabel = geofinder.TKHelper.CLabel(self.root, text="GEO FINDER", width=30, style='Large.TLabel')

        self.original_entry: geofinder.TKHelper.CLabel = geofinder.TKHelper.CLabel(self.root, text="   ", width=GFStyle.TXT_WID, style='Light.TLabel')
        self.user_entry: geofinder.TKHelper.CEntry = geofinder.TKHelper.CEntry(self.root, text="   ", width=GFStyle.TXT_WID, font=(GFStyle.FNT_NAME, 14))
        self.status: geofinder.TKHelper.CLabel = geofinder.TKHelper.CLabel(self.root, width=GFStyle.TXT_WID, style='Good.TLabel')
        self.prefix: geofinder.TKHelper.CLabel = geofinder.TKHelper.CLabel(self.root, width=GFStyle.TXT_WID, style='Highlight.TLabel')

        # User fix statistics
        self.statistics_text: geofinder.TKHelper.CLabel = geofinder.TKHelper.CLabel(self.root, text="",
                                                                                    width=GFStyle.TXT_WID,
                                                                                  style='Light.TLabel')

        # Treeview (list box)
        self.tree_scrollbar = ttk.Scrollbar(self.root)

        self.tree = ttk.Treeview(self.root, style="Plain.Treeview", selectmode="browse")
        self.tree.tag_configure('odd', background=GFStyle.ODD_ROW_COLOR)
        self.tree.tag_configure('even', background='white')

        self.tree["columns"] = ("pre",)
        self.tree.column("#0", width=500, minwidth=100, stretch=tk.NO)
        self.tree.column("pre", width=180, minwidth=5, stretch=tk.NO)
        self.tree.heading("#0", text="   Location", anchor=tk.W)
        self.tree.heading("pre", text="   Prefix", anchor=tk.W)

        self.tree.config(yscrollcommand=self.tree_scrollbar.set)
        self.tree_scrollbar.config(command=self.tree.yview)

        self.ged_event_info: geofinder.TKHelper.CLabel = geofinder.TKHelper.CLabel(self.root, text=" ", width=GFStyle.TXT_WID, style='Light.TLabel')
        self.footnote: geofinder.TKHelper.CLabel = geofinder.TKHelper.CLabel(self.root, text="Data is from GeoNames.org.  Hover for details",
                                                                             width=GFStyle.TXT_WID, style='Light.TLabel')

        self.prog: Progress.Progress = Progress.Progress(self.root, bar_color=GFStyle.HIGH_COLOR, trough_color=GFStyle.LT_GRAY, status=self.status)

        self.search_button: ttk.Button = ttk.Button(self.root, text="search", command=self.main.search_handler,
                                                    width=GFStyle.BTN_WID, image=self.images['search'], compound="left")
        self.map_button: ttk.Button = ttk.Button(self.root, text="map", command=self.main.map_handler,
                                                 width=GFStyle.BTN_WID, image=self.images['map'], compound="left")
        self.verify_button: ttk.Button = ttk.Button(self.root, text="verify", command=self.main.verify_handler,
                                                    width=GFStyle.BTN_WID, image=self.images['verify'], compound="left")
        self.save_button: ttk.Button = ttk.Button(self.root, text="save", command=self.main.save_handler,
                                                  width=GFStyle.BTN_WID, image=self.images['save'], compound="left")
        self.skip_button: ttk.Button = ttk.Button(self.root, text="skip", command=self.main.skip_handler, width=GFStyle.BTN_WID,
                                                  image=self.images['skip'], compound="left")
        self.help_button: ttk.Button = ttk.Button(self.root, text="help", command=self.main.help_handler,
                                                  width=GFStyle.BTN_WID, image=self.images['help'], compound="left")
        self.quit_button: ttk.Button = ttk.Button(self.root, text=" quit", command=self.main.quit_handler,
                                                  width=GFStyle.BTN_WID, image=self.images['exit'], compound="left")

        # Set grid layout for padding column widget - just pads out left column
        self.pad.grid(column=PAD_COL, row=0, padx=GFStyle.PAD_PADX, pady=0, sticky="EW")

        # Set grid layout for column 0 (TXT_COL) Widgets
        # The first 8 are set span to 2 columns because the listbox has a scrollbar next to it
        self.title.grid(column=TXT_COL, row=0, padx=0, pady=12, sticky="N", columnspan=2)
        self.prog.bar.grid(column=TXT_COL, row=1, padx=0, pady=5, sticky="EW", columnspan=2)
        self.original_entry.grid(column=TXT_COL, row=2, padx=0, pady=5, sticky="EWS", columnspan=2)
        self.user_entry.grid(column=TXT_COL, row=3, padx=0, pady=0, sticky="EWN", columnspan=2)
        self.status.grid(column=TXT_COL, row=4, padx=0, pady=0, sticky="EW", columnspan=2)
        self.tree.grid(column=TXT_COL, row=5, padx=0, pady=5, sticky="EW")
        self.ged_event_info.grid(column=TXT_COL, row=7, padx=0, pady=6, sticky="W", columnspan=2)
        self.footnote.grid(column=TXT_COL, row=12, padx=0, pady=5, sticky="EW", columnspan=2)
        self.statistics_text.grid(column=TXT_COL, row=11, padx=0, pady=5, sticky='EW', columnspan=2)


        # Column 1 - just the scrollbar
        self.tree_scrollbar.grid(column=SCRL_COL, row=5, padx=0, pady=5, sticky='WNS')

        # Column 2 Widgets
        self.map_button.grid(column=BTN_COL, row=1, padx=GFStyle.BTN_PADX, pady=6, sticky="NE")
        self.search_button.grid(column=BTN_COL, row=2, padx=GFStyle.BTN_PADX, pady=6, sticky="E")
        self.verify_button.grid(column=BTN_COL, row=4, padx=GFStyle.BTN_PADX, pady=6, sticky="E")
        self.save_button.grid(column=BTN_COL, row=5, padx=GFStyle.BTN_PADX, pady=6, sticky="NE")

        self.skip_button.grid(column=BTN_COL, row=6, padx=GFStyle.BTN_PADX, pady=6, sticky="E")
        self.help_button.grid(column=BTN_COL, row=8, padx=GFStyle.BTN_PADX, pady=5, sticky="E")
        self.quit_button.grid(column=BTN_COL, row=12, padx=GFStyle.BTN_PADX, pady=6, sticky="SE")

        # Set accelerator keys for Verify, listbox, and Save
        self.user_entry.bind("<Return>", self.main.return_key_event_handler)
        self.user_entry.bind("<Control-s>", self.main.ctl_s_event_handler)

        # Track whether user is in Edit box or list box
        self.user_entry.bind("<FocusIn>", self.main.entry_focus_event_handler)
        self.tree.bind("<FocusIn>", self.main.list_focus_event_handler)

        # Tooltips
        footnote_text = 'This uses data from GeoNames.org. This work is licensed under a Creative Commons Attribution 4.0 License,\
         see https://creativecommons.org/licenses/by/4.0/ The Data is provided "as is" without warranty or any representation of accuracy, \
         timeliness or completeness.'
        Tooltip.Tooltip(self.root, self.footnote, text=footnote_text)
        Tooltip.Tooltip(self.root, self.ged_event_info, text="Person and Event in GEDCOM")
        Tooltip.Tooltip(self.root, self.original_entry, text="Original GEDCOM entry")
        Tooltip.Tooltip(self.root, self.verify_button, text="Verify new entry")
        Tooltip.Tooltip(self.root, self.search_button, text="Bring up search in browser")
        Tooltip.Tooltip(self.root, self.map_button, text="Bring up map in browser")
        Tooltip.Tooltip(self.root, self.save_button, text="Save this replacement")
        Tooltip.Tooltip(self.root, self.skip_button, text="Ignore this error and write out unmodified")

        self.review_buttons: List[ttk.Button] = [self.save_button, self.search_button, self.verify_button,
                                                 self.skip_button, self.map_button, self.help_button]

        self.root.update()





#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

import csv
import glob
import logging
import os
import time
from collections import namedtuple
from typing import Dict

from geofinder import CachedDictionary, Country, GeoDB, GeoKeys, Loc, AlternateNames


class GeodataFiles:
    """
    Read in geonames.org geo data files, filter them and place the entries in sqlite db.

    The files must be downloaded from geonames.org and used according to their usage rules.
    
    geoname_data - Any of the following: gb.TXT (or other country), allCountries.txt, cities500.txt, etc.

    The geoname file is filtered to only include the countries and feature codes specified in the country_list
    dictionary (stored in a pickle file) and the feature_list dictionary (also from a pickle)
         
     There is a separate UtilMain.py utility which allows users to:
         1. edit country list config file.  Only entries for those countries will be read.    "country_list.pkl"
         2. edit the geoname feature list.  Only those geonames features are included.  "features.pkl"
         3. if files or the above are changed, the utility deletes the DB
    """

    def __init__(self, directory: str, progress_bar):
        self.logger = logging.getLogger(__name__)
        self.geodb = None
        self.directory: str = directory
        self.progress_bar = progress_bar
        self.line_num = 0
        self.cache_changed: bool = False
        sub_dir = GeoKeys.get_cache_directory(self.directory)
        self.country = None

        # Read in dictionary listing Geoname features we should include
        self.feature_code_list_cd = CachedDictionary.CachedDictionary(sub_dir, "feature_list.pkl")
        self.feature_code_list_cd.read()
        self.feature_code_list_dct: Dict[str, str] = self.feature_code_list_cd.dict

        # Read in dictionary listing countries (ISO2) we should include
        self.supported_countries_cd = CachedDictionary.CachedDictionary(sub_dir, "country_list.pkl")
        self.supported_countries_cd.read()
        self.supported_countries_dct: Dict[str, str] = self.supported_countries_cd.dict

        # Read in dictionary listing languages (ISO2) we should include
        self.languages_list_cd = CachedDictionary.CachedDictionary(sub_dir, "languages_list.pkl")
        self.languages_list_cd.read()
        self.languages_list_dct: Dict[str, str] = self.languages_list_cd.dict
        self.lang_list = []

        for item in self.languages_list_dct:
            self.lang_list.append(item)

        self.entry_place = Loc.Loc()

        # Support for Geonames AlternateNames file.  Adds alternate names for entries
        self.alternate_names = AlternateNames.AlternateNames(directory_name=self.directory,
                                                             geo_files=self, progress_bar=self.progress_bar,
                                                             filename='alternateNamesV2.txt', lang_list=self.lang_list)

    def read(self) -> bool:
        """
        .. module:: read
        :synopsis: Read in dictionary for supported countries and features
        :returns: True if error
        """
        # Read list of Supported countries and features - only these countries/features will be loaded from geoname data
        err = self.feature_code_list_cd.read()
        self.feature_code_list_dct = self.feature_code_list_cd.dict
        if err:
            self.logger.error('Error reading features list')
            return True

        err = self.supported_countries_cd.read()
        self.supported_countries_dct = self.supported_countries_cd.dict
        if err:
            self.logger.error('Error reading  supported countries')
            return True

        self.logger.debug('done loading geodata')

        return False

    def read_geoname(self) -> bool:
        # Read Geoname DB file - this is the db of geoname.org city files and is stored in cache directory under geonames_data
        # The db only contains important fields and only for supported countries
        # This file is much smaller and faster to read than the geoname files
        # If the db doesn't exist, read the geonames.org files and build it.
        # the UtilMain.py allows user changes to config parameters and then requires rebuild of db
        # if the user loads a new geonames.org file, we also need to rebuild the db

        # Use db if it exists and is newer than the geonames directory
        cache_dir = GeoKeys.get_cache_directory(self.directory)
        fullpath = os.path.join(cache_dir, 'geodata.db')
        self.logger.debug(f'read_geo db: {fullpath}')

        if os.path.exists(fullpath):
            # See if db exists and is fresh (newer than other files)
            dir_time = os.path.getmtime(self.directory)
            cache_time = os.path.getmtime(fullpath)
            if cache_time > dir_time:
                self.logger.info(f'db up to date: {fullpath}')
                self.geodb = GeoDB.GeoDB(os.path.join(cache_dir, 'geodata.db'))
                # Ensure DB has reasonable number of records
                count = self.geodb.get_row_count()
                self.logger.info(f'Geoname entries = {count:,}')
                if count > 1000:
                    # No error if DB has over 1000 records
                    return False

        # DB  is stale or not available, rebuild it from geoname files
        self.geodb = GeoDB.GeoDB(os.path.join(cache_dir, 'geodata.db'))
        self.country = Country.Country(self.progress_bar, geodb=self.geodb, lang_list=self.lang_list)

        # walk thru list of files ending in .txt e.g US.txt, FR.txt, all_countries.txt, etc
        self.logger.debug(f'{fullpath} not new.  Building db ')
        file_count = 0

        # Clear out all geo_data data since we are rebuilding
        self.geodb.clear_geoname_data()

        # Put in country data
        self.country.read()

        start_time = time.time()

        # Put in geonames file data
        for fname in glob.glob(os.path.join(self.directory, "*.txt")):
            # Read all geoname files except the  utility files and add to db
            if os.path.basename(fname) not in ["admin2Codes.txt", "admin1CodesASCII.txt",
                                               "alternateNamesV2.txt", "alternateNames.txt"]:
                error = self.read_geoname_file(fname)  # Read in info (lat/long) for all places from

                if error:
                    self.logger.error(f'Error reading geoname file')
                else:
                    file_count += 1

        if file_count == 0:
            self.logger.error(f'No geonames files found in {os.path.join(self.directory, "*.txt")}')
            return True

        # Put in alias names
        self.logger.info(f'geonames files done.  Elapsed ={time.time() - start_time}')

        start_time = time.time()
        self.alternate_names.read()
        self.logger.info(f'Alternate names done.  Elapsed ={time.time() - start_time}')
        self.logger.info(f'Geonames entries = {self.geodb.get_row_count():,}')

        start_time = time.time()
        self.progress("Create Indices", 95)
        self.geodb.create_geoid_index()
        self.geodb.create_indices()
        self.logger.debug(f'Indices done.  Elapsed ={time.time() - start_time}')

        return False

    def read_geoname_file(self, file) -> bool:  # , g_dict
        """Read in geonames files and build lookup structure

        Read a geoname.org places file and create a db of all the places.
        1. The db contains: Name, Lat, Long, district1ID (State or Province ID),
        district2_id, feat_code

        2. Since Geonames supports over 25M entries, the db is filtered to only the countries and feature types we want
        """
        Geofile_row = namedtuple('Geofile_row',
                                 'id name name_asc alt lat lon feat_class feat_code iso iso2 admin1_id'
                                 ' admin2_id admin3_id admin4_id pop elev dem timezone mod')
        self.line_num = 0
        self.progress("Reading {}...".format(file), 0)
        path = os.path.join(self.directory, file)

        if os.path.exists(path):
            fsize = os.path.getsize(path)
            bytes_per_line = 128
            with open(path, 'r', newline="", encoding='utf-8', errors='replace') as geofile:
                self.progress("Loading {}".format(file), 2)  # initialize progress bar
                reader = csv.reader(geofile, delimiter='\t')
                self.geodb.db.begin()

                # Map line from csv reader into GeonameData namedtuple
                for line in reader:
                    self.line_num += 1
                    if self.line_num % 80000 == 0:
                        # Periodically update progress
                        self.progress(msg="Loading {}".format(file), val=self.line_num * bytes_per_line * 100 / fsize)
                    try:
                        geoname_row = Geofile_row._make(line)
                    except TypeError:
                        self.logger.error(f'Unable to read Geoname file {file}.  Line {self.line_num}')
                        continue

                    # Only handle line if it's  for a country we follow and its
                    # for a Feature tag we're interested in
                    if geoname_row.iso.lower() in self.supported_countries_dct and \
                            geoname_row.feat_code in self.feature_code_list_dct:
                        self.insert_georow(geoname_row)

                    if self.progress_bar is not None:
                        if self.progress_bar.shutdown_requested:
                            # Abort DB build.  Clear out partial DB
                            self.geodb.clear_geoname_data()

            self.progress("Write Database", 90)
            self.geodb.db.commit()
            self.progress("Database created", 100)
            return False
        else:
            return True

    def insert_georow(self, geoname_row):
        # Create Geo_row and inses
        # ('paris', 'fr', '07', '012', 12.345, 45.123, 'PPL', '34124')
        geo_row = [None] * GeoDB.Entry.MAX
        geo_row[GeoDB.Entry.NAME] = GeoKeys.normalize(geoname_row.name)
        geo_row[GeoDB.Entry.SDX] = GeoKeys.get_soundex(geo_row[GeoDB.Entry.NAME])

        geo_row[GeoDB.Entry.ISO] = geoname_row.iso.lower()
        geo_row[GeoDB.Entry.ADM1] = geoname_row.admin1_id
        geo_row[GeoDB.Entry.ADM2] = geoname_row.admin2_id
        geo_row[GeoDB.Entry.LAT] = geoname_row.lat
        geo_row[GeoDB.Entry.LON] = geoname_row.lon
        geo_row[GeoDB.Entry.FEAT] = geoname_row.feat_code
        geo_row[GeoDB.Entry.ID] = geoname_row.id

        if 'bonaf' in geo_row[GeoDB.Entry.NAME]:
            print(f'INS {geo_row[GeoDB.Entry.NAME]}')

        self.geodb.insert(geo_row=geo_row, feat_code=geoname_row.feat_code)

    def get_supported_countries(self) -> [str, int]:
        """ Convert list of supported countries into sorted string """
        # todo implement list of names of supported countries
        nm_msg = ""
        for ky in self.supported_countries_dct:
            nm_msg += ky.upper() + ', '
        return nm_msg, len(self.supported_countries_dct)

    def progress(self, msg, val):
        """ Update progress bar if there is one """
        if val > 80:
            self.logger.info(f'{val:.1f}%  {msg}')
        else:
            self.logger.debug(f'{val:.1f}%  {msg}')

        if self.progress_bar is not None:
            self.progress_bar.update_progress(val, msg)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
import copy
import glob
import logging
import os
import sys
import webbrowser
from pathlib import Path
from tkinter import filedialog
from tkinter import messagebox

from geofinder import Geodata, GeoKeys, Config, Gedcom, Loc, AppLayout, UtilLayout, GrampsXml
from geofinder import __version__
from geofinder.CachedDictionary import CachedDictionary
from geofinder.Geodata import ResultFlags
from geofinder.TKHelper import TKHelper

MISSING_FILES = 'Missing Files.  Please run geoutil and correct errors in Errors Tab'
file_types = 'GEDCOM / Gramps XML'

GEOID_TOKEN = 1
PREFIX_TOKEN = 2


class GeoFinder:
    """
    Read in a GEDCOM or Gramps genealogy file and verify that the spelling of each place is correct.
    If place is found, add the latitude and longitude to the output  file.
    If place can't be found, allow the user to correct it.  If it is corrected, apply the change to
    all matching entries.
    Also allow the user to mark an item to skip if a resolution cant be found.
    Skiplist is a list of locations the user has flagged to skip.  These items will be ignored
    Global replace list is list of fixes the user has found.  Apply these to any new similar matches.
    Accepted list is a list of matches that the user has accepted.
        
    Uses gazetteer files from geoname.org as the reference source.  Uses English modern place names,
    plus some county aliases.

    Main classes:

    GeoFinder - The main GUI
    GeoData - The geonames data model routines
    GeoDB -  Database insert/lookup routines
    GEDCOM - routines to read and write GEDCOM files
    GrampsXML - routines to read and write GrampsXML files
    GeodataFile - routines to read/write geoname data sources
    AppLayout - routines to create the app windows and widgets
    Loc - holds all info for a single location

    """

    def __init__(self):
        print('GeoFinder v{}'.format(__version__.__version__))
        print('Python {}.{}'.format(sys.version_info[0], sys.version_info[1]))

        if sys.version_info < (3, 6, 0):
            raise Exception("GeoFinder Requires Python 3.6 or higher.")
        val = ''
        print(f'GeoFinder Requires Python 3.6 or higher {val}')
        self.save_enabled = False  # Only allow SAVE when we have an item that was matched in geonames
        self.user_selected_list = False  # Indicates whether user selected a list entry or text edit entry
        self.err_count = 0
        self.clean_count = 0
        self.skip_count = 0
        self.odd = False
        self.ancestry_file_handler = None
        self.place = None
        self.skiplist = None
        self.global_replace = None
        self.user_accepted = None
        self.geodata = None
        self.out_suffix = ''

        self.logger = self.setup_logging('geofinder Init')

        # Save our base directory and cache directory
        self.directory: str = os.path.join(str(Path.home()), GeoKeys.get_directory_name())
        self.cache_dir = GeoKeys.get_cache_directory(self.directory)

        # Create App window and configure  window buttons and widgets
        self.w: AppLayout.AppLayout = AppLayout.AppLayout(self)
        self.util = UtilLayout.UtilLayout(root=self.w.root, directory=self.directory, cache_dir=self.cache_dir)
        self.w.create_initialization_widgets()
        self.w.config_button.config(state="normal")

        # Set up configuration  class
        self.cfg = Config.Config()

        # Ensure GeoFinder directory structure is valid
        if self.cfg.valid_directories():
            # Directories are valid.  See if  required Geonames files are present
            err = self.check_configuration()
            if err:
                # Missing files
                self.logger.warning('Missing files')
                self.w.status.set_text("Click Config to set up Geo Finder")
                TKHelper.set_preferred_button(self.w.config_button, self.w.initialization_buttons, "Preferred.TButton")
                self.w.load_button.config(state="disabled")
            else:
                # No config errors
                # Read config settings (Ancetry file path)
                err = self.cfg.read()
                if err:
                    self.logger.warning('error reading {} config.pkl'.format(self.cache_dir))

                self.w.original_entry.set_text(self.cfg.get("gedcom_path"))
                TKHelper.enable_buttons(self.w.initialization_buttons)
                if os.path.exists(self.cfg.get("gedcom_path")):
                    #  file is valid.  Prompt user to click Open for  file
                    self.w.status.set_text(f"Click Open to load {file_types} file")
                    TKHelper.set_preferred_button(self.w.load_button, self.w.initialization_buttons, "Preferred.TButton")
                else:
                    # No file.  prompt user to select a  file - GEDCOM file name isn't valid
                    self.w.status.set_text(f"Choose a {file_types} file")
                    self.w.load_button.config(state="disabled")
                    TKHelper.set_preferred_button(self.w.choose_button, self.w.initialization_buttons, "Preferred.TButton")
        else:
            # Missing directories
            self.logger.warning('Directories not found: {} '.format(self.cache_dir))
            self.w.status.set_text("Click Config to set up Geo Finder")
            self.w.load_button.config(state="disabled")
            TKHelper.set_preferred_button(self.w.config_button, self.w.initialization_buttons, "Preferred.TButton")

            # Create directories for GeoFinder
            self.cfg.create_directories()

        # Flag to indicate whether we are in startup or in Window loop.  Determines how window idle is called
        self.startup = False
        self.w.root.mainloop()  # ENTER MAIN LOOP and Wait for user to click on load button

    def load_data(self):
        # Read in Skiplist, Replace list and  Already Accepted list
        self.skiplist = CachedDictionary(self.cache_dir, "skiplist.pkl")
        self.skiplist.read()
        self.global_replace = CachedDictionary(self.cache_dir, "global_replace.pkl")
        self.global_replace.read()
        self.user_accepted = CachedDictionary(self.cache_dir, "accepted.pkl")  # List of items that have been accepted
        self.user_accepted.read()

        # Initialize geodata
        self.geodata = Geodata.Geodata(directory_name=self.directory, progress_bar=self.w.prog)
        error = self.geodata.read()
        if error:
            TKHelper.fatal_error(MISSING_FILES)

        # If the list of supported countries is unusually short, display note to user
        num = self.display_country_note()
        self.logger.info('{} countries will be loaded'.format(num))

        # Read in Geoname Gazeteer file - city names, lat/long, etc.
        error = self.geodata.read_geonames()
        if error:
            TKHelper.fatal_error(MISSING_FILES)
        self.logger.debug(f'Geoname dictionary has {self.geodata.geo_files.geodb.get_row_count()} entries')
        self.w.root.update()
        self.w.prog.update_progress(100, " ")

    def load_handler(self):
        """
        User pressed LOAD button to load an Ancestry file. Switch app display to the Review Widgets
        Load in file name and
        loop through  file and find every PLACE entry and verify the entry against the geoname data
        """
        self.w.original_entry.set_text("")
        self.w.remove_initialization_widgets()  # Remove old widgets
        self.w.create_review_widgets()  # Switch display from Initial widgets to main review widgets

        self.load_data()

        ged_path = self.cfg.get("gedcom_path")  # Get saved config setting for  file

        # Depending on file suffix, load appropriate handler
        if ged_path is not None:
            if '.ged' in ged_path:
                self.out_suffix = "new.ged"
                self.ancestry_file_handler = Gedcom.Gedcom(in_path=ged_path, out_suffix=self.out_suffix, cache_d=self.cache_dir,
                                                           progress=self.w.prog)  # Routines to open and parse GEDCOM file
            elif '.gramps' in ged_path:
                self.out_suffix = "new.gramps"
                self.ancestry_file_handler = GrampsXml.GrampsXml(in_path=ged_path, out_suffix=self.out_suffix, cache_d=self.cache_dir,
                                                                 progress=self.w.prog)  # Routines to open and parse Gramps file
        else:
            self.out_suffix = 'unk.new.ged'

        if self.ancestry_file_handler.error:
            TKHelper.fatal_error(f"File {ged_path} not found.")

        self.w.root.update()

        self.place: Loc.Loc = Loc.Loc()  # Create an object to store info for the current Place

        # Add  filename to Title
        path_parts = os.path.split(ged_path)  # Extract filename from full path
        self.w.title.set_text(f'GEO FINDER - {path_parts[1]}')

        # Read  file, find each place entry and handle it.
        self.handle_place_entry()

    def get_replacement(self, dct, town_entry: str, place):
        geoid = None
        replacement = dct.get(town_entry)

        if replacement is not None:
            if len(replacement) > 0:
                # parse replacement entry
                rep_tokens = replacement.split('@')
                geoid = rep_tokens[GEOID_TOKEN]
                if len(geoid) > 0:
                    self.geodata.find_geoid(geoid, place)
                else:
                    place.result_type = GeoKeys.Result.DELETE

                # Get prefix if there was one
                if len(rep_tokens) > 2:
                    place.prefix = rep_tokens[PREFIX_TOKEN]

        return geoid

    def handle_place_entry(self):
        """ Get next PLACE  in users  File.  Replace it, skip it, or have user correct it. """
        self.w.original_entry.set_text("")

        if self.w.prog.shutdown_requested:
            self.periodic_update("Shutting down...")
        else:
            self.periodic_update("Scanning")
        self.clear_detail_text(self.place)

        while True:
            self.err_count += 1  # Counter is used to periodically update
            # Update statistics
            self.w.statistics_text.set_text(f'Done={self.clean_count}   Skipped={self.skip_count}   '
                                            f'Remaining={self.ancestry_file_handler.place_total - (self.clean_count + self.skip_count)}')

            # Find the next PLACE entry in  file
            # Process it and keep looping until we need user input
            self.place.clear()
            town_entry, eof = self.ancestry_file_handler.get_next_place()

            if eof:
                self.end_of_file_shutdown()

            # When we find an error, see if we already have a fix (Global Replace) or Skip (ignore).
            # Otherwise have user handle it
            geoid = self.get_replacement(self.global_replace, town_entry, self.place)

            if geoid is not None:
                # There is a global change that we can apply to this line.
                self.clean_count += 1

                if self.place.result_type == GeoKeys.Result.EXACT_MATCH:
                    # Output updated place to ancestry file
                    self.write_updated_place(self.place)

                    # Display status to user
                    if self.w.prog.shutdown_requested:
                        self.periodic_update("Shutting down...")
                    else:
                        self.periodic_update("Applying change")
                elif self.place.result_type == GeoKeys.Result.DELETE:
                    continue
                else:
                    self.logger.debug(f'Error looking up GEOID=[{geoid}] for [{town_entry}] ')
                    self.place.event_year = int(self.ancestry_file_handler.event_year)  # Set place date to event date (geo names change over time)
                    self.geodata.find_location(town_entry, self.place)
                    break

                continue
            elif self.get_replacement(self.user_accepted, town_entry, self.place) is not None:
                # There is an accepted  change that we can use for this line.  Get the replacement text
                self.clean_count += 1

                if self.place.result_type == GeoKeys.Result.EXACT_MATCH:
                    # Output place to  file
                    self.write_updated_place(self.place)

                    # Display status to user
                    if self.w.prog.shutdown_requested:
                        self.periodic_update("Shutting down...")
                    else:
                        self.periodic_update("Applying change")
                else:
                    self.logger.debug(f'Error looking up GEOID {town_entry}')
                    self.place.event_year = int(self.ancestry_file_handler.event_year)  # Set place date to event date (geo names change over time)
                    self.geodata.find_location(town_entry, self.place)
                    self.w.original_entry.set_text(self.place.name)  # Display place
                    self.w.user_entry.set_text(self.place.name)  # Display place
                    break

                continue
            elif self.w.prog.shutdown_requested:
                # User requested shutdown.  Finish up going thru file, then shut down
                self.periodic_update("Shutting Down...")
                self.w.original_entry.set_text(" ")
                self.ancestry_file_handler.write_asis()
                continue
            elif self.skiplist.get(town_entry) is not None:
                # item is in skiplist - Write out as-is and go to next error
                # self.logger.debug(f'Found Skip for {town_entry}')
                self.skip_count += 1

                self.periodic_update("Skipping")
                self.ancestry_file_handler.write_asis()
                continue
            else:
                # Found a new PLACE entry
                # See if it is in our place database
                # self.place.parse_place(place_name=town_entry, geo_files=self.geodata.geo_files)
                self.place.event_year = int(self.ancestry_file_handler.event_year)  # Set place date to event date (geo names change over time)
                self.geodata.find_location(town_entry, self.place)

                if self.place.result_type in GeoKeys.successful_match:
                    # Found a match
                    self.clean_count += 1
                    if self.place.result_type == GeoKeys.Result.EXACT_MATCH:
                        # Exact match
                        # Write out line without user verification
                        if self.w.prog.shutdown_requested:
                            self.periodic_update("Shutting down...")
                        else:
                            self.periodic_update("Scanning")

                        # Add to global replace list - Use '@' for tokenizing.  Save GEOID_TOKEN and PREFIX_TOKEN
                        res = '@' + self.place.geoid + '@' + self.place.prefix

                        self.global_replace.set(town_entry, res)
                        self.logger.debug(f'Found Exact Match for {town_entry} res= [{res}] Setting DICT')
                        # Periodically flush dictionary to disk.  (We flush on exit as well)
                        if self.err_count % 4 == 1:
                            self.global_replace.write()

                        self.write_updated_place(self.place)
                        continue
                    else:
                        # Have user review the match
                        self.w.status.configure(style="Good.TLabel")
                        self.w.original_entry.set_text(self.place.name)  # Display place
                        self.w.user_entry.set_text(self.place.name)  # Display place
                        break
                else:
                    self.logger.debug(f'User2 review for {town_entry}')

                    self.w.status.configure(style="Good.TLabel")
                    self.w.original_entry.set_text(self.place.name)  # Display place
                    self.w.user_entry.set_text(self.place.name)  # Display place
                    # Have user review the match
                    break

        # Have user review the result
        self.display_result(self.place)

    def get_list_selection(self):
        # Get the item the user selected in list (tree)
        loc = (self.w.tree.item(self.w.tree.selection(), "text"))
        prefix = (self.w.tree.item(self.w.tree.selection())['values'][0])
        return prefix, loc

    def verify_handler(self):
        """ The User clicked verify.  Verify if the users new Place entry has a match in geonames data.  """
        # Do we verify item from listbox or from text edit field?
        if self.user_selected_list:
            # User selected item from listbox - get listbox selection
            pref, town_entry = self.get_list_selection()
            self.logger.debug(f'first match {town_entry}')

            # Update the user edit widget with the List selection item
            # self.w.user_entry.set_text(town_entry)

            # Since we are verifying users choice, Get first match.  don't try partial match
            self.geodata.find_first_match(town_entry, self.place)
            self.place.prefix = pref
        else:
            # User typed in text entry window - get edit field value and look it up
            town_entry: str = self.w.user_entry.get()
            if len(town_entry) > 0:
                self.geodata.find_location(town_entry, self.place)
            else:
                # User entry is blank - prompt to delete this entry
                self.place.clear()
                self.place.result_type = GeoKeys.Result.DELETE
                self.logger.debug('Blank: DELETE')
                self.geodata.process_result(self.place, '', ResultFlags(limited=False, filtered=False))

        self.display_result(self.place)

    def display_result(self, place):
        """ Display result details for an item  """
        # Enable buttons so user can either click Skip, or edit the item and Click Verify.
        TKHelper.enable_buttons(self.w.review_buttons)
        nm = place.format_full_name()
        self.logger.debug(f'disp result [{place.prefix}][{place.prefix_commas}][{nm}] res=[{place.result_type}]')

        # Enable action buttons based on type of result
        if place.result_type == GeoKeys.Result.MULTIPLE_MATCHES or \
                place.result_type == GeoKeys.Result.NO_MATCH or \
                place.result_type == GeoKeys.Result.NO_COUNTRY:
            # Disable the Save & Map button until user clicks Verify and item is found
            self.set_save_allowed(False)
            TKHelper.set_preferred_button(self.w.verify_button, self.w.review_buttons, "Preferred.TButton")
        elif place.result_type == GeoKeys.Result.NOT_SUPPORTED:
            # Found a match or Not supported - enable save and verify
            #self.set_save_allowed(True)  # Enable save button
            TKHelper.set_preferred_button(self.w.skip_button, self.w.review_buttons, "Preferred.TButton")
        else:
            # Found a match or Not supported - enable save and verify
            self.set_save_allowed(True)  # Enable save button
            TKHelper.set_preferred_button(self.w.save_button, self.w.review_buttons, "Preferred.TButton")

        # Display status and color based on success
        self.set_status_text(place.get_status())
        if place.result_type in GeoKeys.successful_match:
            if place.place_type == Loc.PlaceType.CITY:
                self.w.status.configure(style="Good.TLabel")
            else:
                self.w.status.configure(style="GoodCounty.TLabel")
        else:
            self.w.status.configure(style="Error.TLabel")

        if len(place.georow_list) > 0:
            # Display matches in listbox
            self.w.tree.focus()  # Set focus to listbox
            self.display_georow_list(place)
        else:
            # No matches
            self.w.user_entry.focus()  # Set focus to text edit widget
            self.display_one_georow(place.status_detail)

        # Display GEDCOM person and event that this location refers to
        self.w.ged_event_info.set_text(
            f'{self.ancestry_file_handler.get_name(self.ancestry_file_handler.id)}: '
            f'{self.ancestry_file_handler.event_name} {self.ancestry_file_handler.date}')
        self.w.root.update_idletasks()

    def list_insert(self, text, prefix):
        # Tags to alternate colors in lists
        self.odd = not self.odd
        if self.odd:
            tag = ('odd',)
        else:
            tag = ('even',)
        self.w.tree.insert(parent='', index="end", iid=None, text=text, values=(prefix,), tags=tag)

    def clear_display_list(self):
        self.odd = False
        for row in self.w.tree.get_children():
            self.w.tree.delete(row)

    def display_georow_list(self, place: Loc.Loc):
        """ Display list of matches in listbox (tree) """

        # Clear listbox
        self.clear_display_list()

        temp_place = copy.copy(place)

        # Get geodata for each item and add to listbox output
        for geo_row in place.georow_list:
            self.geodata.geo_files.geodb.copy_georow_to_place(geo_row, temp_place)
            nm = temp_place.format_full_name()
            valid = self.geodata.validate_year_for_location(event_year=place.event_year, iso=temp_place.country_iso,
                                                            admin1=temp_place.admin1_id, padding=0)
            if valid:
                self.list_insert(nm, temp_place.prefix)
            else:
                self.list_insert(nm, "INVALID DATE")

        self.w.root.update_idletasks()

    def skip_handler(self):
        """ Write out original entry as-is and skip any matches in future  """
        self.skip_count += 1
        self.logger.debug(f'Skip for {self.w.original_entry.get_text()}  Updating SKIP dict')

        self.skiplist.set(self.w.original_entry.get_text(), " ")
        self.ancestry_file_handler.write_asis()

        # Save Skip info for future runs
        self.skiplist.write()

        # Go to next entry
        self.handle_place_entry()

    def save_handler(self):
        """ Save the Place.  Add Place to global replace list and replace if we see it again. """
        self.clean_count += 1

        ky = self.w.original_entry.get_text()
        if self.place.result_type == GeoKeys.Result.DELETE:
            # Put in a blank as replacement
            self.place.geoid = ''
            self.place.prefix = ''

        res = '@' + self.place.geoid + '@' + self.place.prefix
        self.logger.debug(f'Save [{ky}] :: [{res}]')

        # Add item to global replace list if user made a change.  This will be cached to disk.
        if self.w.original_entry.get_text() != self.w.user_entry.get_text():
            # User made a change - save it
            self.global_replace.set(ky, res)
            # Periodically flush dict to disk
            if self.err_count % 3 == 1:
                self.global_replace.write()
            self.logger.debug(f'SAVE SetGblRep for [{ky}] res=[{res}] Updating DICT')
        else:
            # User accepted the item as is.  Add to accept list
            self.logger.debug(f'SAVE Accept for [{ky}] res=[{res}] Updating DICT')

            self.user_accepted.set(ky, res)
            self.user_accepted.write()

        # Write out corrected item to  output file
        if self.place.result_type != GeoKeys.Result.DELETE:
            self.write_updated_place(self.place)

        # Get next error
        self.w.user_entry.set_text('')
        self.handle_place_entry()

    @staticmethod
    def help_handler():
        """ Bring up browser with help text """
        help_base = "https://github.com/corb555/GeoFinder/wiki/User-Guide"
        webbrowser.open(help_base)

    def search_handler(self):
        """ Bring up browser with Google search with this item """
        base = "https://www.google.com/search?q="
        loc = self.w.user_entry.get_text()
        webbrowser.open(f"{base}{loc}")

    def map_handler(self):
        """ Bring up browser with map for this item """
        base = "http://www.openstreetmap.org"
        if self.place.place_type == Loc.PlaceType.COUNTRY or self.place.place_type == Loc.PlaceType.ADMIN1:
            # Zoom wide if user just put in state, country
            zoom = "zoom=7"
        else:
            zoom = "zoom=14"
        loc = f"{base}/?mlat={self.place.lat}&mlon={self.place.lon}&{zoom}#map={self.place.lat}/{self.place.lon}"
        webbrowser.open(loc)

    def quit_handler(self):
        """ Set flag for shutdown.  Process all global replaces and exit """
        path = self.cfg.get("gedcom_path")
        self.w.prog.shutdown_requested = True

        if messagebox.askyesno(' ', f'  Updates saved. Do you want to save output file?\n\n {path}.{self.out_suffix}'):
            TKHelper.disable_buttons(button_list=self.w.review_buttons)
            self.w.quit_button.config(state="disabled")
            self.w.prog.startup = True

            self.w.user_entry.set_text(" ")
            self.w.status.set_text("Quitting...")
            # Write out the item we are on
            self.ancestry_file_handler.write_asis()

            # We will still continue to go through file, but only handle global replaces
            self.handle_place_entry()
        else:
            self.logger.info('Shutdown')
            self.shutdown()

    def config_handler(self):
        # User clicked on Config button - bring up configuration windows
        self.w.remove_initialization_widgets()  # Remove old widgets
        self.w.root.protocol("WM_DELETE_WINDOW", self.shutdown)
        # Read config settings (Ancestry file path) - this will create config.pkl if not found
        err = self.cfg.read()
        if err:
            self.logger.warning('error reading {} config.pkl'.format(self.cache_dir))
        self.util.create_util_widgets()  # Create Util/Config widgets

    def filename_handler(self):
        """ Display file open selector dialog """
        fname = filedialog.askopenfilename(initialdir=self.directory,
                                           title=f"Select {file_types} file",
                                           filetypes=[("GEDCOM files", "*.ged"), ("Gramps files", "*.gramps"), ("all files", "*.*")])
        if len(fname) > 1:
            self.cfg.set("gedcom_path", fname)  # Add filename to dict
            self.cfg.write()  # Write out config file
            self.w.status.set_text(f"Click Open to load {file_types} file")
            self.w.original_entry.set_text(self.cfg.get("gedcom_path"))
            TKHelper.set_preferred_button(self.w.load_button, self.w.initialization_buttons, "Preferred.TButton")

    def return_key_event_handler(self, _):
        """ User pressed Return accelerator key.  Call Verify data entry """
        self.verify_handler()
        return "break"

    def entry_focus_event_handler(self, _):
        # Track focus so when user presses Verify, we know whether to get text from Entry Box or List box
        # User clicked on data entry widget
        # Note - second param of _ is to prevent warning for Event param which isnt used
        self.user_selected_list = False

    def list_focus_event_handler(self, _):
        # Track focus so when user presses Verify, we know whether to get text from Entry Box or List box
        # user clicked on listbox
        self.user_selected_list = True

    def ctl_s_event_handler(self, _):
        """ User pressed Ctrl-S Save accelerator key.  Call Save  """
        if self.save_enabled:
            self.save_handler()
        return "break"

    def clear_detail_text(self, place):
        self.odd = False
        self.clear_display_list()
        place.georow_list.clear()

    def display_one_georow(self, txt):
        self.logger.debug(f'DISP ONE ROW {txt}')
        self.clear_display_list()
        # self.w.scrollbar.grid_remove()  # Just one item, so hide scrollbar
        self.list_insert(txt, '')

    def display_country_note(self) -> int:
        """ display warning if only a small number of countries are enabled """
        country_list, supported_countries = self.geodata.geo_files.get_supported_countries()
        self.w.root.update()
        if supported_countries == 0:
            TKHelper.fatal_error("No countries enabled.\n\nUse Config Country Tab to change country list\n")

        # if supported_countries < 3:
        #    messagebox.showinfo("Info", f"Loading geocode data for the following ISO country codes:"
        #                                f"\n\n{country_list}\n\nUse Config Country Tab to change country list\n")
        return supported_countries

    @staticmethod
    def setup_logging(msg):
        logger = logging.getLogger(__name__)
        fmt = "%(levelname)s %(name)s.%(funcName)s %(lineno)d: %(message)s"
        logging.basicConfig(level=logging.DEBUG, format=fmt)
        logger.info(msg)
        return logger

    def write_updated_place(self, place: Loc.Loc):
        # Write out updated location and lat/lon to  file
        nm = place.format_full_name()
        # self.logger.debug(f'ged write [{place.prefix}][{place.prefix_commas}][{nm}]')
        if place.result_type != GeoKeys.Result.DELETE:
            self.ancestry_file_handler.write_updated(place.prefix + place.prefix_commas + nm)
            self.ancestry_file_handler.write_lat_lon(lat=place.lat, lon=place.lon)
        else:
            self.logger.debug('zero len, no output')

    def shutdown(self):
        """ Shutdown - write out Gbl Replace and skip file and exit """
        # self.w.root.update_idletasks()
        if self.skiplist:
            self.skiplist.write()
        if self.global_replace:
            self.global_replace.write()
        if self.user_accepted:
            self.user_accepted.write()
        if self.cfg:
            self.cfg.write()
        if self.ancestry_file_handler:
            self.ancestry_file_handler.close()
        self.w.root.quit()
        self.logger.info('SYS EXIT')
        # sys.exit()
        os._exit(0)

    def set_save_allowed(self, save_allowed: bool):
        if save_allowed:
            # Enable the Save and Map buttons
            self.save_enabled = True
            self.w.save_button.config(state="normal")  # Match - enable  button
            self.w.map_button.config(state="normal")  # Match - enable  button
        else:
            # Disable the Save and Map buttons
            self.save_enabled = False
            self.w.save_button.config(state="disabled")
            self.w.map_button.config(state="disabled")

    def set_status_text(self, txt):
        # status text is readonly
        self.w.status.state(["!readonly"])
        self.w.status.set_text(txt)
        self.w.status.state(["readonly"])

    def end_of_file_shutdown(self):
        # End of file reached
        TKHelper.disable_buttons(button_list=self.w.review_buttons)
        self.w.status.set_text("Done.  Shutting Down...")
        self.w.original_entry.set_text(" ")
        path = self.cfg.get("gedcom_path")
        messagebox.showinfo("Info", f"Finished file.\n\nOutput file:\n {path}.{self.out_suffix}")
        self.logger.info('End of  file')
        self.shutdown()

    def periodic_update(self, msg):
        # Display status to user
        if self.err_count % 30 == 0:
            self.w.status.set_text(msg)
            self.w.status.configure(style="Good.TLabel")
            self.w.original_entry.set_text(self.place.name)  # Display place
            self.w.root.update_idletasks()  # Let GUI update

    def check_configuration(self):
        file_error = False
        file_list = ['allCountries.txt', 'cities500.txt']

        # Get country list and validate
        countries = CachedDictionary(self.cache_dir, 'country_list.pkl')
        self.logger.debug('load country selections list')

        countries.read()
        country_dct = countries.dict
        if len(country_dct) == 0:
            self.logger.warning('no countries specified')
            file_error = True

        # Ensure that there is geo DB or there are some geoname data files
        path = os.path.join(self.directory, self.cache_dir, "geodata.db")
        if os.path.exists(path):
            self.logger.info(f'Found {path}')
            file_error = False
            return file_error

        path = os.path.join(self.directory, "*.txt")
        self.logger.info(f'Geoname path {path}')
        count = 0

        country_file_len = 6
        for filepath in glob.glob(path):
            # Ignore the two Admin files
            fname = os.path.basename(filepath)
            if len(fname) == country_file_len or fname in file_list:
                count += 1

        self.logger.debug(f'geoname file count={count}')
        if count == 0:
            # No data files, add error to error dictionary
            self.logger.warning('No Geonames files found')
            file_error = True

        return file_error


def entry():
    GeoFinder()


if __name__ == "__main__":
    entry()
